#!/usr/bin/env python
import argparse
from smurfs._smurfs.smurfs import Smurfs
from smurfs._smurfs.multi_smurfs import MultiSmurfs
from smurfs.__version__ import __version__
from smurfs.support.mprint import *
import os
import pickle
import subprocess
import sys

def main(args = None):
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser()

    parser.add_argument("target", help="Target for the frequency analysis. Can be either a target, or "
                                       "a star name. If it is a target, it needs to be an ASCII target. "
                                       "The first two rows are expected to be time "
                                       "and flux, the third row is optional and can be used as the flux "
                                       "error.", type=str)
    parser.add_argument("snr", help="The desired lower bound for the signal to noise ratio for the analysis",
                        type=float)
    parser.add_argument("windowSize", help="The window size used to get the SNR for a given frequency", type=float)
    parser.add_argument("-fr", "--frequencyRange",
                        help="Optional parameter that describes the desired frequency range. "
                             "The application will only check within this parameters if given. "
                             "Per default this parameter is set from 0 to Nyquist frequency."
                             "Please enter it in a form of a tuple (lower,higher).",
                        type=str, default="None,None")

    parser.add_argument("-ssa", "--skipSimilarFrequencies", help="If this parameter is set, the frequencies surrounding"
                                                                 "one frequency are in a very similar range, that area "
                                                                 "will be ignored", action='store_true')

    parser.add_argument("-sc", "--skipCutoff",
                        help="Skips the cutoff check, so similar frequencies don't chancel the run. Be aware that this "
                             "might lead to unknown behaviour",
                        action='store_true')

    parser.add_argument("-ef","--extendFrequencies",
                        help="Extends the frequency analysis by n frequencies, meaning, the analysis will only stop"
                             "after n insignificant frequencies are found.",type=int,default=0)

    parser.add_argument("-imf","--improveFitMode",help="Three different fit improvement modes are available:\n1)'all' "
                                                       "-> improve fit after every frequency\n2)'end' -> improve "
                                                       "fit after last significant frequency was found.\n3) 'none' -> "
                                                       "Disables improve fit.",type=str,choices=['all','end','none'],
                        default='all')

    parser.add_argument("-fm", "--fitMethod", help="Using this flag you can either choose 'scipy' "
                                                   "(scipy.optimize.curve_fit) or 'lmfit' (lmfti Model fit). "
                                                   "'lmfit' is activated by default",
                        type=str, choices=["scipy", "lmfit"], default="lmfit")

    parser.add_argument("-ft","--fluxType", help="If you input SC Tess observations (TIC IDs), this flag "
                                                 "allows you to choose between PDCSAP and SAP flux. Default "
                                                 "is PDCSAP",
                        type=str,choices=["PDCSAP","SAP"],default="PDCSAP")

    parser.add_argument("-so","--storeObject",help="If this flag is set, the SMURFS object is stored in the "
                                                   "results. You can later use this object to load the "
                                                   "_result into a python file using 'Smurfs.from_path'."
                        ,action='store_true')

    parser.add_argument("-sp","--savePath",help="Allows you to set the save path of the analysis. By "
                                                "default it will save it in the same folder, where "
                                                "the module was called.",type=str,default=".")

    parser.add_argument("-i", "--interactive", help="Using this flag will automatically start an iPython shell. "
                                                    "This allows for direct interaction with the result using the "
                                                    "'star' object."
                        , action='store_true')



    """
    #todo replace this
    parser.add_argument("-om", "--outputMode", help="Optional parameter that defines what will be exported as dataset. "
                                                    "There are two possibilities: Normal and Full. Normal mode means, that"
                                                    "only the first and last amplitudespectra will be plotted & exported, "
                                                    "full means all amplitudespectra will be plotted & exported. Default is "
                                                    "Normal.",
    
                        type=str, choices=["Normal", "Full"], default="Normal")
                        
    parser.add_argument("-igr", "--ignoreCutoffRatio", help="Optional parameter. If this is set to True, it will ignore"
                                                            "the gap ratio cutoff criterion", action='store_true')  
                                                            
    parser.add_argument("-trs", "--timeBaseSplit",
                        help="Optional parameter that describes the split within the time range."
                             "This can be used to get multiple amplitude spectra for the "
                             "lightcurve, split up within the time Range. A parameter could be "
                             "for example 50, where the parameter would split the lightcurve in"
                             "50 day chunks (last segment will be smaller/equal). Defaults"
                             "to -1, which means no splitting will occur",
                        type=int, default=-1)
    parser.add_argument("-o", "--overlap",
                        help="Optional parameter that describes the overlap of two split ranges. Will only"
                             "work, if -trs is set to a reasonable parameter. For example, if the trs is "
                             "set to 50 and the overlap to 2, the chunks would contain the ranges (0,50),"
                             "(48,98),(96,146) ... Defaults to 0, which means no overlap will take place",
                        type=int, default=0)                                                                                  
    """
    parser.add_argument("--version", help="Shows version of SMURFS", action='version',
                        version='SMURFS {version}'.format(version=__version__))



    args = parser.parse_args(args)

    if "," in args.target:
        targets = args.target.split(",")
    else:
        targets = [args.target]

    f_min = None if args.frequencyRange.split(",")[0] == 'None' else float(args.frequencyRange.split(",")[0])
    f_max = None if args.frequencyRange.split(",")[1] == 'None' else float(args.frequencyRange.split(",")[1])

    if len(targets) == 1:
        target = targets[0]

        if len(target.split(".")) == 2 and os.path.basename(target).split(".")[1] in ['txt','dat']:
            s = Smurfs(file=target)
        else:
            s = Smurfs(target_name=target,flux=args.fluxType)

        improve_fit = args.improveFitMode == 'all'

        s.run(snr=args.snr,window_size=args.windowSize,f_min=f_min,f_max=f_max,
              skip_similar=args.skipSimilarFrequencies,similar_chancel=not args.skipCutoff
              ,extend_frequencies=args.extendFrequencies,improve_fit=improve_fit
              ,mode=args.fitMethod)

        if args.improveFitMode == 'end':
            mprint("Improving fit ...", state)
            s.improve_result()

        s.save(args.savePath,args.storeObject)

        #start interactive shell, by loading smurfs object directly into the shell
        if args.interactive:
            mprint("Starting interactive shell. Use the 'star' object to interact with the result!",state)
            pickle.dump(s, open("i_obj.smurfs", "wb"))
            cmd = ["ipython","-i","-c","from smurfs import Smurfs;import pickle;star : Smurfs = pickle.load(open('i_obj.smurfs', 'rb'));import os;os.remove('i_obj.smurfs');import matplotlib.pyplot as pl;pl.ion()"]
            subprocess.call(cmd)
            mprint("Done", info)
    else:
        if len(targets[0].split(".")) == 2 and os.path.basename(targets[0]).split(".")[1] in ['txt','dat']:
            s = MultiSmurfs(file_list=targets)
        else:
            s = MultiSmurfs(target_list=targets,flux_types=args.fluxType)

        s.run(snr=args.snr,window_size=args.windowSize,f_min=f_min,f_max=f_max,
              skip_similar=args.skipSimilarFrequencies,similar_chancel=not args.skipCutoff
              ,extend_frequencies=args.extendFrequencies,improve_fit=args.disableImproveFrequencies
              ,mode=args.fitMethod)
        s.save(args.savePath, args.storeObject)

if __name__ == "__main__":
    main()


from lightkurve import search_lightcurvefile, TessLightCurve, TessLightCurveFile
from typing import List,Tuple
from astroquery.simbad import Simbad
from astropy.coordinates import SkyCoord
import astropy.units as u
import re
import eleanor
from astroquery.mast import Tesscut
from astroquery.mast import Catalogs
import numpy as np
import astropy.units as u
from smurfs.support.mprint import *
from contextlib import redirect_stdout
import io
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.gridspec import GridSpec
import matplotlib.pyplot as pl


def aperture_contour(ax : Axes, obj : eleanor.TargetData) -> Axes:
    """
    Plots a countour plot of the TPF with the aperture by Eleanor.
    :param ax: Axes object where this is plotted on
    :param obj: TargetData object containing TPF
    :return: used axes object
    """
    aperture = obj.aperture

    ax.imshow(obj.tpf[0])

    f = lambda x, y: aperture[int(y), int(x)]
    g = np.vectorize(f)

    x = np.linspace(0, aperture.shape[1], aperture.shape[1] * 100)
    y = np.linspace(0, aperture.shape[0], aperture.shape[0] * 100)
    X, Y = np.meshgrid(x[:-1], y[:-1])
    Z = g(X[:-1], Y[:-1])

    ax.contour(Z[::-1], [0.5], colors='w', linewidths=[4],
                extent=[0 - 0.5, x[:-1].max() - 0.5, 0 - 0.5, y[:-1].max() - 0.5])

def create_validation_page(data_list: List[eleanor.TargetData], q_list: List[np.ndarray],
                           title : str) -> Figure:
    """
    Creates a validation page from the eleanor extraction of light curves from TESS FFIs.
    Contains the pixel cutout, the original and reduced light curve.
    :param data_list: List of tpfs from eleanor
    :param q_list: Quality flag list
    :param title: Title of pplot
    """
    fig: Figure = pl.figure(figsize=(14, 12 + 4 * len(data_list)))
    height_ratio = [2] + [1 for i in range(len(data_list) - 1)] + [1]
    gs = GridSpec(len(data_list) + 1, 4, width_ratios=[1, 1, 1, 1], height_ratios=height_ratio)
    ax: Axes = fig.add_subplot(gs[0, :])

    for data, q in zip(data_list, q_list):
        ax.plot(data.time[q], data.corr_flux[q] - np.median(data.corr_flux[q]), '.', markersize=2,
                label=f"Sector {data.source_info.sector}")
    ax.legend()
    ax.set_xlabel("Time")
    ax.set_ylabel("Flux")

    for i, (data, q) in enumerate(zip(data_list, q_list)):
        ax: Axes = fig.add_subplot(gs[i + 1, 0])
        aperture_contour(ax, data)
        ax.set_title("TPF with aperture")

        ax: Axes = fig.add_subplot(gs[i + 1, 1:])
        ax.plot(data.time[q],
                data.raw_flux[q] / np.median(data.raw_flux[q]) - 0.01, 'k.',
                label='Raw flux')
        ax.plot(data.time[q],
                data.corr_flux[q] / np.median(data.corr_flux[q]) + 0.01, 'r.',
                label='Corrected flux')
        ax.legend()

        ax.set_title(f"Sector {data.source_info.sector}")

    fig.suptitle(f"{title}")
    fig.subplots_adjust(hspace=0.3)

    return fig


def mag(lc: TessLightCurve) -> TessLightCurve:
    """
    Converts and normalizes a LighCurve object to magnitudes.
    :param lc: lightcurve object
    :return: reduced light curve object
    """
    lc = lc.remove_nans()
    lc.flux = -2.5 * np.log10(lc.flux)
    lc = lc.remove_nans()
    lc.flux -= np.median(lc.flux)
    lc.flux = lc.flux * u.mag
    return lc


def cut_ffi(tic_id: int) -> Tuple[TessLightCurve,Figure]:
    """
    Extracts light curves from FFIs using TESScut and Eleanor. Querying all available sectors for a given TIC ID.
    :param tic_id: ID of the star in the TIC
    :return: Tess light curve and figure containing validation page
    """
    f = io.StringIO()
    with redirect_stdout(f):
        try:
            stars = eleanor.multi_sectors(tic=tic_id, sectors='all')
        except:
            stars = eleanor.multi_sectors(tic=tic_id, sectors='all', tc=True)
    mprint(f.getvalue().strip(), log)

    lc_list = []
    data_list = []
    q_list = []

    for star in stars:
        data = eleanor.TargetData(star, height=15, width=15, bkg_size=31, do_psf=False, do_pca=False)
        q = data.quality == 0
        lc_list.append(TessLightCurve(time=data.time[q], flux=data.corr_flux[q], targetid=tic_id))

        data_list.append(data)
        q_list.append(q)

    fig = create_validation_page(data_list,q_list,f'TIC {tic_id}')

    lc: TessLightCurve = lc_list[0]
    lc = mag(lc)
    if len(lc_list) > 1:
        for l in lc_list:
            lc = lc.append(mag(l))
    mprint(f"Extracted light curve for TIC {tic_id}!", info)
    return lc, fig


def get_tess_ffi_lc(target_name: str) -> Tuple[TessLightCurve,Figure]:
    """
    Tries to extract a light curve from TESS FFIs. If you do not provide a TIC number, it will try
    to look up the target on Simbad and link it to the TESS catalog using Simbad.
    :param target_name: Name of the target. It is recommended to use the TIC ID directly, if you provide
    any other name, it needs to be able to be looked up in Simbad. It then links the target by the coordinates
    from Simbad to the TIC catalogue.
    :return: A TessLightCurve object containing the Light curve of the object, validation page figure
    """
    if target_name.startswith('TIC'):
        tic_id = re.findall(r'\d+', target_name)
        if len(tic_id) == 0:
            raise ValueError(ctext("A Tic ID needs to consist of TIC and a number!", error))
        tic_id = int(tic_id[0])
    else:
        mprint(f"Querying simbad for {target_name}", log)

        simbad_query = Simbad.query_object(target_name)
        if simbad_query is None:
            raise ValueError(ctext(f"Can't find a simbad entry for {target_name}", error))

        c = SkyCoord(simbad_query[0]['RA'] + " " + simbad_query[0]['DEC'], unit=(u.hourangle, u.deg))
        mprint(f"Found {target_name} at {c}. Searching for TESS observations using TESScut ...", info)

        hdulist = Tesscut.get_sectors(coordinates=c)
        if len(hdulist) == 0:
            raise ValueError(ctext(f"Can't find a TESS observation for {target_name}", error))
        try:
            tic_star = Catalogs.query_region(f"{c.ra} {c.dec}", radius=0.0001, catalog="TIC")
        except:
            tic_star = Catalogs.query_region(f"{c.ra.value} {c.dec.value}", radius=0.0001, catalog="TIC")

        if len(tic_star) > 1:
            raise RuntimeWarning(ctext("Found two stars in the coordinates of this star!", error))

        tic_star = tic_star[0]
        tic_id = tic_star['ID']
        mprint(f"Found TESS observations for {target_name} with TIC {tic_id}", info)

    mprint(f"Extracting light curves from FFIs, this may take a bit ... ", log)
    return cut_ffi(tic_id)


def download_lc(target_name: str, flux_type='PDCSAP') -> Tuple[TessLightCurve,Figure]:
    """
    Downloads a light curve using the TESS mission. If the star has been observed in the SC mode, it
    will download the original light curve from MAST. You can also choose the flux type you want to use.

    If it wasn't observed in SC mode, it will try to extract a light curve from the FFIs if the target has
    been observed by TESS.
    :param target_name: The name of the target. It is recommended to use the TIC ID directly, if you provide
    any other name, it needs to be able to be looked up in Simbad. It then links the target by the coordinates
    from Simbad to the TIC catalogue.
    :param flux_type: Type of flux in the SC mode. Can be either PDCSAP or SAP
    :return: A TessLightCurve object containing the Light curve of the objectl, validation page if
    extracted from FFI
    """
    mprint(f'Searching processed light curves for {target_name} ...', log)
    res = search_lightcurvefile(target_name)
    if len(res) == 0:
        mprint(f'No light curve found using light kurve. Checking TESS FFIs ...', log)
        lc,fig = get_tess_ffi_lc(target_name)
    else:
        fig = None
        mprint(f"Found processed light curve for {target_name}!", info)
        res = res.download_all()
        type = 'TESS' if isinstance(res.data[0], TessLightCurveFile) else 'Kepler'
        mprint(f"Using {type} observations! Combining sectors ...", log)

        if flux_type == 'PDCSAP':
            lc: TessLightCurve = res.PDCSAP_FLUX.data[0]
        elif flux_type == 'SAP':
            lc: TessLightCurve = res.SAP_FLUX.data[0]
        else:
            raise ValueError(ctext("Flux type needs to be either PDCSAP or SAP", error))

        lc = mag(lc)
        if len(res) > 1:
            data: List[TessLightCurve] = res.PDCSAP_FLUX.data[1:] if flux_type == 'PDCSAP' else res.SAP_FLUX.data[1:]
            for d in data:
                lc = lc.append(mag(d))

    lc = lc.remove_nans()
    lc = lc.remove_outliers(4)
    mprint(f"Total observation length: {'%.2f' % (lc.time[-1] - lc.time[0])} days.", log)
    return lc,fig



import os
import numpy as np
from lightkurve import LightCurve
from smurfs.preprocess.tess import mag
from smurfs.support.mprint import *

def load_file(file : str) -> LightCurve:
    """
    Loads and normalizes target content
    :param file: Name of target including path
    :return: LightCurve object
    """
    if not os.path.exists(file):
        raise IOError(ctext(f"File {file} doesn't exist!",error))

    mprint(f"Reading data from {file} ...",log)
    data = np.loadtxt(file)
    if data.shape[0] > data.shape[1]:
        data = data.T

    if data.shape[0] == 2:
        lc = LightCurve(time=data[0],flux=data[1])
    else:
        lc = LightCurve(time=data[0], flux=data[1],flux_err=data[2])

    lc = lc.remove_nans()
    lc.flux = lc.flux + float(np.amin(lc.flux)) + 10
    lc = mag(lc)
    lc = lc.remove_outliers(4)
    lc = lc.remove_nans()
    mprint(f"Total observation length: {'%.2f' % (lc.time[-1] - lc.time[0])} days.",log)
    mprint("Extracted data from target!",info)
    return lc